// SimpleCloth.cpp: implementation of the SimpleCloth class.
//
//////////////////////////////////////////////////////////////////////

#include "Parameter.h"
#include "SimpleCloth.h"
#include "glUtilities.h"
#ifndef __AFXWIN_H__
#include <afxwin.h>
#endif
#include <GL/gl.h>
#include <math.h>
#include <stdlib.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SimpleCloth::SimpleCloth()
{
	myParameters.Add(Parameter("cDamping",0.0,5.0,0.15));
	myParameters.Add(Parameter("density",0.05,5.0,0.3));
	myParameters.Add(Parameter("gravity",-20.0,20.0,-9.8));
	myParameters.Add(Parameter("kBend",0.0,2000.0,150.0));
	myParameters.Add(Parameter("kShear",0.0,8000.0,4000.0));
	myParameters.Add(Parameter("kStretch",0.0,16000.0,8000.0));
	myParameters.Add(Parameter("windMinimum",-500.0,500.0,40.0));
	myParameters.Add(Parameter("windPressure",-500.0,500.0,20.0));
	myParameters.Add(Parameter("windFrequency",0.0,50.0,6.0));

	mayaParam.outfile= NULL;
	mayaParam.frame_no= 0;

	orig= curr= prev= next= NULL;
	kShear= 4000.0;					// N/m
	kStretch= 8000.0;				// N/m
	kBend= 150.0;					// N/m
	density= 0.3;					// kg/m^2
	gravity.GetElement()[1]= -9.8;	// m/s^2
	cDamping= 0.15;					// kg/s
	windMinimum= 40.0;				// N/m^2 (Pa)
	windPressure= 20.0;				// N/m^2 (Pa)
	windFrequency= 6;				// rad/s
	windDirection.GetElement()[2]= 1.0;
	windPhase= 0.0;
	windForce= windDirection*windPressure;
	srand((unsigned int)time(NULL));	// seed the random number generator
	coordSystem.SetDimension(4,4);
}

SimpleCloth::~SimpleCloth()
{
	FreeMem();
	if(mayaParam.outfile!= NULL)
		MayaEndRecording();
}

void SimpleCloth::FreeMem()
{
	int i;

	if(orig != NULL) {
		for(i=nX*nY-1; i>=0; i--)
			delete orig[i];
		delete [] orig;
		orig= NULL;
	}
	if(curr != NULL) {
		for(i=nX*nY-1; i>=0; i--)
			delete curr[i];
		delete [] curr;
		curr= NULL;
	}
	if(prev != NULL) {
		for(i=nX*nY-1; i>=0; i--)
			delete prev[i];
		delete [] prev;
		prev= NULL;
	}
	if(next != NULL) {
		for(i=nX*nY-1; i>=0; i--)
			delete next[i];
		delete [] next;
		next= NULL;
	}
}

int SimpleCloth::Create(double space,int nx,int ny,Matrix *CSystem)
{
	int index,ix,iy;
	Vector src(4),pos(4);

	coordSystem.GetElement()[15]= 1.0;
	if(CSystem==NULL) {
		coordSystem.GetElement()[0]= 1.0;	// local-x aligned with world-x
		coordSystem.GetElement()[5]= -1.0;	// local-y aligned with neg. world-y
		coordSystem.GetElement()[10]= -1.0;	// local-z aligned with neg. world-z
	}
	else
		coordSystem= *CSystem;

	FreeMem();
	if((space < 0.0) || (nx*ny < 3))
		return -1;

	nX= nx;
	nY= ny;
	spacing= space;
	orig= new Vector*[nx*ny];
	curr= new Vector*[nx*ny];
	prev= new Vector*[nx*ny];
	next= new Vector*[nx*ny];
	if((curr == NULL) || (prev == NULL) || (next == NULL))
		return -1;

	src.GetElement()[3]= 1.0;
	for(iy= 0; iy<nY; iy++) {
		src.GetElement()[1]= spacing*iy;
		for(ix= 0; ix<nX; ix++) {
			src.GetElement()[0]= spacing*ix;
			pos= coordSystem*src;
			index= iy*nX + ix;
			orig[index]= new Vector(3,pos.GetElement());
			curr[index]= new Vector(3,pos.GetElement());
			prev[index]= new Vector(3,pos.GetElement());
			next[index]= new Vector(3,pos.GetElement());
		}
	}

	return 0;
}

double SimpleCloth::ShearK(double k)
{
	double temp= kShear;

	if(k<0.0)
		return kShear;
	kShear= k;

	return temp;
}

double SimpleCloth::StretchK(double k)
{
	double temp= kStretch;

	if(k<0.0)
		return kStretch;
	kStretch= k;

	return temp;
}

double SimpleCloth::BendK(double k)
{
	double temp= kBend;

	if(k<0.0)
		return kBend;
	kBend= k;

	return temp;
}

double SimpleCloth::Density(double rho)
{
	double temp= density;

	if(rho<0.0)
		return density;
	density= rho;

	return temp;
}

double SimpleCloth::DampingC(double c)
{
	double temp= cDamping;

	if(c<0.0)
		return cDamping;
	cDamping= c;

	return temp;
}

void SimpleCloth::glDraw()
{
	int ix,iy;
	Vector normal(3);

	if(!Ready())
		return;

	for(iy= 0; iy<nY-1; iy++) {
		glBegin(GL_QUAD_STRIP);
		for(ix= 0; ix<nX; ix++) {
			glNormal3dv(Normal(curr,ix,iy).GetElement());
			glVertex3dv(curr[iy*nX+ix]->GetElement());
			glNormal3dv(Normal(curr,ix,iy+1).GetElement());
			glVertex3dv(curr[iy*nX+ix+nX]->GetElement());
		}
		glEnd();
	}
}

int SimpleCloth::MayaStartRecording(const char *fullpath,int num_u,int num_v)
{
	char filename[250];
	char path[250];
	char name[50];
	char *spos;

	strcpy(path,fullpath);
	spos= strrchr(path,'\\');	// look for end of path
	strcpy(mayaParam.name,spos+1);		// extract entity name
	mayaParam.num_u= num_u;
	mayaParam.num_v= num_v;
	if(spos != NULL)
		*(spos+1)= (char)0;		// terminate the new path
	else
		*path= (char)0;			// empty string
	spos= strchr(name,'.');		// look for extension
	if(spos != NULL)
		*spos= (char)0;			// terminate name string
	
// Output the MEL commands to create a NURBS surface with the given number of u and v patches.
// nurbsPlane -p 0 0 0 -ax 0 1 0 -w 30 -lr 4 -d 3 -u 5 -v 20 -ch 1; objectMoveCommand;
	sprintf(filename,"%screate_%s.mel",path,mayaParam.name);
	mayaParam.outfile= fopen(filename,"w");
	if(mayaParam.outfile == NULL)
		return -1;
	fprintf(mayaParam.outfile,"nurbsPlane -p 0 0 0");
	fprintf(mayaParam.outfile," -ax %1.3f %1.3f %1.3f", coordSystem.GetElement()[2], coordSystem.GetElement()[6], coordSystem.GetElement()[10] );
	fprintf(mayaParam.outfile," -w %1.3f", spacing*nX );
	fprintf(mayaParam.outfile," -lr %1.3f", (float)nY/nX );
	fprintf(mayaParam.outfile," -d %d", 3 );
	fprintf(mayaParam.outfile," -u %d", mayaParam.num_u );
	fprintf(mayaParam.outfile," -v %d;\n", mayaParam.num_v );
	fprintf(mayaParam.outfile,"rename \"%s\";\n",mayaParam.name);
	fclose(mayaParam.outfile);

	sprintf(filename,"%sanimate_%s.mel",path,mayaParam.name);
	mayaParam.outfile= fopen(filename,"w");

	return mayaParam.outfile == NULL;
}

int SimpleCloth::MayaRecordFrame(int frame)
{
	if(mayaParam.outfile == NULL)
		return -1;
	if(frame < 1)
		mayaParam.frame_no++;
	else
		mayaParam.frame_no= frame;
// Output the MEL commands to keyframe the results of the simulation.
	fprintf(mayaParam.outfile,"currentTime %d;\n",mayaParam.frame_no);
	if(WriteFrame(mayaParam.outfile,mayaParam.name,mayaParam.num_u,mayaParam.num_v))
		return -1;
	fprintf(mayaParam.outfile,"setKeyframe -shape 1 {\"%s.cv[0:%d][0:%d]\"};\n",mayaParam.name,mayaParam.num_u+2,mayaParam.num_v+2);

	return 0;
}

int SimpleCloth::MayaEndRecording()
{
	if(mayaParam.outfile != NULL)
		fclose(mayaParam.outfile);
	mayaParam.outfile= NULL;

	return 0;
}

int SimpleCloth::WriteFrame(FILE *outfile,const char *name,int num_u,int num_v)
{
// select -r nurbsPlane1.cv[0][0] ;
// setAttr "nurbsPlaneShape1.controlPoints[10].yValue" 1;	int u,v;
	int ui,vi;
	double u,v;
	Vector x;

	for(vi=0; vi<num_v+3; vi++) {
		for(ui=0; ui<num_u+3; ui++) {
			if((ui > 1) && (ui < num_u+1))
				u= (double)(ui-1)/num_u;
			else if(ui == 1)
				u= 1.0/(3*num_u);
			else if(ui == num_u+1)
				u= 1.0-(1.0/(3*num_u));
			else if(ui == 0)
				u= 0.0;
			else
				u= 1.0;

			if((vi > 1) && (vi < num_v+1))
				v= (double)(vi-1)/num_v;
			else if(vi == 1)
				v= 1.0/(3*num_v);
			else if(vi == num_v+1)
				v= 1.0-(1.0/(3*num_v));
			else if(vi == 0)
				v= 0.0;
			else
				v= 1.0;

			x= Position(u,v,curr)-Position(u,v,orig);
			fprintf(outfile,"setAttr \"%s.controlPoints[%d].xValue\" %1.3f;\n",name,ui*(num_v+3)+vi,(float)x.GetElement()[0]);
			fprintf(outfile,"setAttr \"%s.controlPoints[%d].yValue\" %1.3f;\n",name,ui*(num_v+3)+vi,(float)x.GetElement()[1]);
			fprintf(outfile,"setAttr \"%s.controlPoints[%d].zValue\" %1.3f;\n",name,ui*(num_v+3)+vi,(float)x.GetElement()[2]);
		}
	}		
	return 0;
}

Vector SimpleCloth::Normal(Vector **pvector,int ix,int iy)
{
	Vector normal;
	int index;

	index= iy*nX+ix;
	if(iy > 0)
		if(ix>0)
			normal= ((*pvector[index-1] - *pvector[index]).Cross(*pvector[index-nX] - *pvector[index])).Normalize();
		else
			normal= ((*pvector[index-nX] - *pvector[index]).Cross(*pvector[index+1] - *pvector[index])).Normalize();
	else
		if(ix>0)
			normal= ((*pvector[index+nX] - *pvector[index]).Cross(*pvector[index-1] - *pvector[index])).Normalize();
		else
			normal= ((*pvector[index+1] - *pvector[index]).Cross(*pvector[index+nX] - *pvector[index])).Normalize();

	return normal;
}

void SimpleCloth::Increment(double dt)
{
	int y_index,index,ix,iy;
	double nodeMass,diagonal;
	Vector **temp,f_ext(3),delta(3);

	if(!Ready())
		return;
	cDamping= myParameters.LookupValue("cDamping");
	density= myParameters.LookupValue("density");
	gravity.GetElement()[1]= myParameters.LookupValue("gravity");
	kBend= myParameters.LookupValue("kBend");
	kShear= myParameters.LookupValue("kShear");
	kStretch= myParameters.LookupValue("kStretch");
	windMinimum= myParameters.LookupValue("windMinimum");
	windPressure= myParameters.LookupValue("windPressure");
	windFrequency= myParameters.LookupValue("windFrequency");

	nodeMass= spacing*spacing*density;
	diagonal= spacing*sqrt(2.0);

	windPhase+= dt*windFrequency;
	windForce= windDirection*((windPressure*cos(windPhase)*(sin(windPhase)>0.5 ? 1.0 : 0.0)+windMinimum)*spacing*spacing);
	if(rand()<RAND_MAX/12) {
		windCenter.GetElement()[0]= spacing*nX*(0.5-(double)rand()/RAND_MAX);
		windCenter.GetElement()[1]= spacing*nY*(0.5-(double)rand()/RAND_MAX);
	}
	int mx,my;
	mx= nX/2;
	my= nY/2;
	for(iy= 0; iy<nY; iy++) {
		y_index= iy*nX;
		for(ix= 0; ix<nX; ix++) {
			index= y_index + ix;
//			if(!((ix == 0 && iy == 0) || (ix == 0 && iy == nY-1) || (ix == nX-1 && iy == 0) || (ix == nX-1 && iy == nY-1)))
//			if(!(iy == 0 && (ix == 0 || ix == 0)))
			if(!(iy == 0 && ix % 8 == 0))
//			if(((ix-mx)*(ix-mx)+(iy-my)*(iy-my)) > 50)
//			if(!(ix > 3 && ix < 12 && iy > 3 && iy < 12))
//			if(!((ix == 0 && iy == 0) || (ix == nX-1 && iy == 0)))
			{
// compute forces acting on current node (internal and external)
/*			0  1  2
			 \ | /
			7--C--3
			 / | \
			6  5  4
*/
				f_ext= gravity;
// Internal forces toward 0, 1, 2:
				if(iy < nY-1) {
					// toward node 0	
					if(ix > 0) {
						delta= *curr[index + nX - 1] - *curr[index];
						delta= (delta - delta.Normalize()*diagonal)*kShear;
						f_ext= f_ext + delta;
					}
					// toward node 1
					delta= *curr[index + nX] - *curr[index];
					delta= (delta - delta.Normalize()*spacing)*kStretch;
					f_ext= f_ext + delta;
					// toward node 2
					if(ix < nX-1) {
						delta= *curr[index + nX + 1] - *curr[index];
						delta= (delta - delta.Normalize()*diagonal)*kShear;
						f_ext= f_ext + delta;
					}
				}
// Internal forces toward 6, 5, 4:
				if(iy > 0) {
					// toward node 6	
					if(ix > 0) {
						delta= *curr[index - nX - 1] - *curr[index];
						delta= (delta - delta.Normalize()*diagonal)*kShear;
						f_ext= f_ext + delta;
					}
					// toward node 5
					delta= *curr[index - nX] - *curr[index];
					delta= (delta - delta.Normalize()*spacing)*kStretch;
					f_ext= f_ext + delta;
					// toward node 4
					if(ix < nX-1) {
						delta= *curr[index - nX + 1] - *curr[index];
						delta= (delta - delta.Normalize()*diagonal)*kShear;
						f_ext= f_ext + delta;
					}
				}
				// toward 7:
				if(ix > 0) {
					delta= *curr[index - 1] - *curr[index];
					delta= (delta - delta.Normalize()*spacing)*kStretch;
					f_ext= f_ext + delta;
				}
				// toward 3:
				if(ix < nX -1) {
					delta= *curr[index + 1] - *curr[index];
					delta= (delta - delta.Normalize()*spacing)*kStretch;
					f_ext= f_ext + delta;
				}
// Internal bend forces:
				if((ix > 0) && (ix < nX-1)) {
					delta= ((*curr[index + 1] + *curr[index - 1])*0.5 - *curr[index])*kBend;
					f_ext= f_ext + delta;
				}
				if((iy > 0) && (iy < nY-1)) {
					delta= ((*curr[index + nX] + *curr[index - nX])*0.5 - *curr[index])*kBend;
					f_ext= f_ext + delta;
				}
// External wind forces:
				double modulation= (*curr[index] - windCenter).Magnitude();
				modulation*= fabs(Normal(curr,ix,iy)*windDirection);
				f_ext= f_ext + windForce * modulation;
// Compute new position:
				*next[index]= f_ext*(dt*dt) + *curr[index]*(2.0*nodeMass + cDamping*dt) - *prev[index]*nodeMass;
				*next[index]= *next[index]*(1.0/(nodeMass + cDamping*dt));
			}
		}
	}

	temp= prev;
	prev= curr;
	curr= next;
	next= temp;
}

Vector SimpleCloth::Position(double u,double v,Vector **source)
{
	double rx,ry;
	int lx,ly;
	Vector mid1,mid2,result;

	if(source == NULL)
		source= curr;
	rx= u*(nX-1);
	lx= (int)floor(rx);
	ry= v*(nY-1);
	ly= (int)floor(ry);
	rx= rx-(double)lx;
	ry= ry-(double)ly;
	if(lx == nX-1) {
		rx= 1.0;
		lx--;
	}
	if(ly == nY-1) {
		ry= 1.0;
		ly--;
	}
	mid1= (*source[ly*nX+lx+1] - *source[ly*nX+lx])*rx + *source[ly*nX+lx];
	mid2= (*source[(ly+1)*nX+lx+1] - *source[(ly+1)*nX+lx])*rx + *source[(ly+1)*nX+lx];
	result= (mid2-mid1)*ry+mid1;

	return result;
}
